import inspect
from functools import reduce
from typing import Any, Dict, List, Optional, Sequence, Union

from .types import ColumnDef, ColumnSpec, IndexFormatter

__all__ = ["Tree"]


class Tree:
    def __init__(self, parent: Optional["Tree"], path: str) -> None:
        self.parent: Union[Tree, None] = parent
        self.depth: int = 0
        if parent and parent.depth is not None:
            self.depth = parent.depth + 1

        self.path: str = path
        self.children: List[Tree] = []
        self._tree_spec: Optional[ColumnSpec] = None
        self.accepts_index_arg = False

    @property
    def tree_spec(self) -> ColumnSpec:
        if not self._tree_spec:
            raise ValueError("No tree spec found")

        return self._tree_spec

    @property
    def name(self) -> Any:
        if not self._tree_spec:
            raise ValueError("No tree spec found")

        if self._tree_spec.name:
            return self._tree_spec.name
        return self.path

    def get_value(self, value: Any, idx: Union[int, None] = None) -> Any:
        formatter = self.tree_spec.formatter
        if formatter:
            if isinstance(formatter, IndexFormatter) and idx is not None and self.accepts_index_arg:
                return formatter(value, index=idx)
            else:
                return formatter(value)  # type: ignore

        return {self.name: value}

    def add_child(self, child_tree: "Tree") -> None:
        self.children.append(child_tree)

    def get_child(self, path: str) -> "Tree":
        paths = path.split(".")

        def _select_child(tree: "Tree", path: str) -> "Tree":
            return next(filter(lambda p: p.path == path, tree.children), tree)

        child = reduce(lambda tree, path: _select_child(tree, path), paths, self)
        return child

    def __repr__(self) -> str:
        base = f"{self.path}\n"

        for child in self.children:
            sep = "\t" * self.depth
            base += f"{sep}{repr(child)}\n"

        return base


def load_tree(data: Dict, path: str, parent: Union[Tree, None] = None, root_name: str = "root") -> Tree:
    if parent is None:
        tree = Tree(parent, root_name)
    else:
        tree = Tree(parent=parent, path=path)

    if isinstance(data, dict):
        for k, v in data.items():
            child = load_tree(v, k, tree)
            tree.add_child(child)

    return tree


def _set(obj: Dict, path: str, value: Any) -> None:
    *split_path, last = path.split(".")
    for bit in split_path:
        new_obj = obj.setdefault(bit, {})
        if new_obj is not None:
            obj = obj.setdefault(bit, {})
        if new_obj is None:
            obj[bit] = {}
    try:
        obj[last] = value
    except Exception as e:
        raise e


def get_keys_nested_dict(dictionary: dict, parent_key: str = "", sep: str = "."):
    """
    Get all keys in a nested dictionary in dot-delimited format.

    Args:
        dictionary (dict): The nested dictionary.
        parent_key (str): The parent key to append for recursive calls (default is an empty string).
        sep (str): The separator for joining keys (default is dot ".").

    Returns:
        set: A set of keys in dot-delimited format.
    """
    keys = []
    for key, value in dictionary.items():
        full_key = f"{parent_key}{sep}{key}" if parent_key else key
        if isinstance(value, dict):
            keys.extend(get_keys_nested_dict(value, full_key, sep=sep))
        elif isinstance(value, list):
            for _, item in enumerate(value):
                if isinstance(item, dict):
                    keys.extend(get_keys_nested_dict(item, f"{full_key}", sep=sep))
                else:
                    keys.append(f"{full_key}")
        else:
            keys.append(full_key)
    return set(keys)


def load_tree_rep(paths: Sequence[ColumnDef]) -> Tree:
    column_defs = [ColumnSpec.from_def(path) for path in paths]

    tree_dict: Dict[str, Any] = {}
    for col_def in column_defs:
        _set(tree_dict, col_def.path, None)
    tree = load_tree(tree_dict, "None", parent=None)

    for col_def in column_defs:
        child = tree.get_child(col_def.path)
        child._tree_spec = col_def
        if col_def.formatter:
            argspec = inspect.getfullargspec(col_def.formatter)
            child.accepts_index_arg = "index" in argspec.args

    return tree



import itertools
from typing import Any, Dict, List, Optional, Sequence, Set, Union

from .tree import Tree, get_keys_nested_dict, load_tree_rep
from .types import ColumnDef, ColumnSpec

__all__ = ["unwind"]


def _unwind_list(data: List, tree: Tree, level_list: List) -> None:
    for item in data:
        item_output_dict = {}
        item_accum: List[dict] = []
        for child in tree.children:
            value: Union[dict, list] = _unwind(item, child)
            if isinstance(value, list):
                if item_accum and len(item_accum) == len(value):  # Handle adjacents
                    item_accum = [{**a, **b} for a, b in zip(item_accum, value)]
                else:
                    item_accum.extend(value)
            else:
                item_output_dict.update(value)
        if len(item_accum) > 0:
            level_list.extend([{**item_output_dict, **item} for item in item_accum])
        else:
            level_list.append(item_output_dict)


def _get_value(data: dict, tree: Tree) -> Any:
    tree_data = data.get(tree.path)
    if isinstance(data, list):
        return [_unwind(d, tree) for d in data]
    if isinstance(tree_data, list):
        return [tree.get_value(x, idx) for idx, x in enumerate(tree_data)]
    return tree.get_value(tree_data)


def _unwind(data: dict, tree: Tree) -> Union[List, List[Dict[str, Any]], Dict[str, Any], dict, List[dict]]:
    if not tree.children:
        return _get_value(data, tree)

    level_list: List[dict] = []
    level_dict: dict = {}

    tree_data = data.get(tree.path)
    tree_data = {} if tree_data is None else tree_data
    if isinstance(tree_data, list):
        _unwind_list(tree_data, tree, level_list)
    else:
        for child in tree.children:
            value = _unwind(tree_data, child)
            if isinstance(value, list):
                level_list.extend(value)
            else:
                level_dict.update(value)

    if len(level_list) == 0:
        return level_dict

    appended = [{**level_dict, **list_value} for list_value in level_list]
    return appended


def __remove_redundant_paths(keys: List[str]):
    # Sort the keys
    sorted_keys = sorted(keys)

    # List to hold the final set of keys
    final_keys = []

    # Iterate through the sorted list of keys
    for i in range(len(sorted_keys)):
        # Ensure this isn't the last element to avoid IndexError
        if i + 1 < len(sorted_keys):
            # Check if the next key starts with the current key followed by a dot (indicating a parent-child relationship)
            if not sorted_keys[i + 1].startswith(sorted_keys[i] + "."):
                final_keys.append(sorted_keys[i])
        else:
            # Always add the last key since it can't be a prefix of any key that follows
            final_keys.append(sorted_keys[i])

    return final_keys


def __get_keys_from_sample(sample: List[Dict[Any, Any]]) -> Set[str]:
    keys: Set[str] = set()
    for item in sample:
        keys.update(get_keys_nested_dict(item))
    return keys


def unwind(
    data: Union[Dict[Any, Any], List[Dict[Any, Any]]],
    columns: Sequence[ColumnDef],
    allow_extra: bool = False,
    infer_length: Optional[int] = 10,
) -> List[Dict[str, Any]]:
    column_defs = [ColumnSpec.from_def(c) for c in columns]

    if allow_extra:
        keys = get_keys_nested_dict(data) if isinstance(data, dict) else __get_keys_from_sample(data[:infer_length])
        specified_keys = [k[0] if isinstance(k, tuple) else k for k in columns]
        extra_paths = [k for k in keys if k not in specified_keys]
        extra_defs: List[ColumnSpec] = [ColumnSpec(path=k, name=k) for k in extra_paths]
        column_defs.extend(extra_defs)
        path_list = [c.path for c in column_defs]
        unique_paths = __remove_redundant_paths(path_list)

        column_defs = [c for c in column_defs if c.path in unique_paths]

    tree = load_tree_rep(column_defs)

    def apply_unwind(d: Dict, tree: Tree) -> List[Dict[str, Any]]:
        unwound = _unwind({"root": d}, tree)
        if not isinstance(unwound, list):
            return [unwound]
        return unwound

    if isinstance(data, dict):
        return apply_unwind(data, tree)
    else:
        return list(itertools.chain(*[apply_unwind(d, tree) for d in data]))



import sys
from dataclasses import dataclass, field
from typing import Any, Callable, Optional, Protocol, Tuple, TypedDict, Union, runtime_checkable

if sys.version_info < (3, 12):
    from typing_extensions import TypedDict  # noqa: F811
else:
    from typing import TypedDict

__all__ = ["ColumnDef", "ColumnFormatterType", "ColumnSpecDict"]


@runtime_checkable
class IndexFormatter(Protocol):
    def __call__(self, data: Any, index: int = ...):
        ...  # pragma: no cover


ColumnFormatterType = Union[IndexFormatter, Callable[[Any], Any], Callable[[Any, int], Any]]


class ColumnSpecDict(TypedDict, total=False):
    name: str
    formatter: ColumnFormatterType


@dataclass
class ColumnSpec:
    path: str
    name: Optional[str] = field(default=None)
    formatter: Optional[ColumnFormatterType] = field(default=None)

    @classmethod
    def from_tuple(cls, column_def: Tuple[str, ColumnSpecDict]) -> "ColumnSpec":
        path, spec = column_def
        return cls(path, **spec)

    @classmethod
    def from_str(cls, path: str) -> "ColumnSpec":
        return cls(path)

    @classmethod
    def from_def(cls, column_def: "ColumnDef") -> "ColumnSpec":
        if isinstance(column_def, str):
            return cls.from_str(column_def)

        elif isinstance(column_def, tuple):
            return cls.from_tuple(column_def)
        return column_def


ColumnDef = Union[Tuple[str, ColumnSpecDict], str, ColumnSpec]
